<?php

/**
 * Copyright (C) VasylTech, LLC <vasyl@vasyltech.com>
 *
 * This program is commercial software: you are not allowed to redistribute it
 * and/or modify. Unauthorized copying of this file, via any medium is strictly
 * prohibited.
 * For any questions or concerns contact Vasyl Martyniuk <vasyl@vasyltech.com>
 */

/**
 * AAM Plus Package Base Manager
 *
 * @package AAM
 * @author Vasyl Martyniuk <vasyl@vasyltech.com>
 */
class AAM_PlusPackage_Base {

    /**
     * Instance of itself
     * 
     * @var AAM_PlusPackage_Backend 
     * 
     * @access private
     */
    private static $_instance = null;
    
    /**
     * Skip filtering
     * 
     * @var bool
     * 
     * @access protected 
     */
    protected $skipFilter = false;

    /**
     * Undocumented variable
     *
     * @var array
     */
    protected $cache = array();
    
    /**
     * Initialize the extension
     * 
     * @return void
     * 
     * @access protected
     */
    protected function __construct() {
        $backend  = AAM::api()->getConfig('core.settings.backendAccessControl', true);
        $frontend = AAM::api()->getConfig('core.settings.frontendAccessControl', true);
        $api      = AAM::api()->getConfig('core.settings.apiAccessControl', true);
        
        // Capabilities control
        add_action('registered_post_type', array($this, 'registeredPostType'), 999, 2);
        add_filter('aam-map-meta-caps-filter', array($this, 'mapMetaCaps'), 999, 3);
        
        add_filter('aam-post-list-options-filter', array($this, 'addPostListOptions'));
        add_filter('aam-post-where-clause-filter', array(AAM_PlusPackage_Visibility_QueryBuilder::bootstrap(), 'execute'), 10, 3);
        add_action('aam-visibility-initialize-action', array(AAM_PlusPackage_Visibility_Initializer::bootstrap(), 'execute'));
        add_filter('aam-policy-post-resource-action-filter', array($this, 'postActionFilter'));
        
        if ($backend || $frontend || $api) {
            //terms & post restriction handlers
            add_filter('get_terms', array($this, 'filterTerms'), 10, 3);
        }

        // Optimize the performance
        add_filter('found_posts', array($this, 'prefetchTerms'), 10, 2);
        
        // manage access to taxonomies
        add_filter('register_taxonomy_args', array($this, 'registerTaxonomyArgs'), 999, 3);
        
        // filter for `can` post object method
        add_filter('aam-post-action-allowed-filter', array($this, 'allowed'), 10, 3);
        
        // manage default category functionality
        // This cover the post type "post" default category
        add_filter('pre_option_default_category', array($this, 'filterDefaultCategory'));
        add_action('save_post', array($this, 'savePost'), 999, 3);
    }

    /**
     * Undocumented function
     *
     * @param [type] $posts
     * @param [type] $wpQuery
     * @return void
     */
    public function prefetchTerms($posts, $wpQuery) {
        $postType = (!empty($wpQuery->query_vars['post_type']) ? $wpQuery->query_vars['post_type'] : 'post');

        $ids = array_map(function($p) {
            $result = null;

            if (is_numeric($p)) {
                $result = intval($p);
            } elseif (is_a($p, 'WP_Post')) {
                $result = $p->ID;
            } elseif(is_array($p) && isset($p['ID'])) {
                $result = $p['ID'];
            }

            return $result;
        }, $wpQuery->posts);

        $terms = $this->fetchTermsByPostIds($ids, $postType);

        if (is_array($terms)) {
            foreach($terms as $term) {
                if (isset($this->cache[$term->object_id])) {
                    $this->cache[$term->object_id][] = $term;
                } else {
                    $this->cache[$term->object_id] = array($term);
                }
            }
        }

        return $posts;
    }

    /**
     * Undocumented function
     *
     * @param [type] $action
     * @return void
     */
    public function postActionFilter($action) {
        switch ($action) {
            case 'listtoothers':
                $action = 'list_others';
                break;

            case 'readbyothers':
                $action = 'read_others';
                break;

            case 'editbyothers':
                $action = 'edit_others';
                break;

            case 'deletebyothers':
                $action = 'delete_others';
                break;

            case 'publishbyothers':
                $action = 'publish_others';
                break;

            default:
                break;
        }

        return $action;
    }
    
    /**
     * Override default category if defined
     * 
     * @param type $category
     * 
     * @return int
     * 
     * @access public
     */
    public function filterDefaultCategory($category) {
        static $skip = false;

        //check if user category is defined
        if ($skip === false) {
            $user = AAM::api()->getUser();

            $skip = true;
            $default = intval($user->getObject('taxonomy', 'post|category')->get('default'));
            $skip = false;
            
            if (empty($default)) {
                $id      = get_current_user_id();
                $default = AAM::api()->getConfig('feature.post.defaultTerm.user.' . $id , null);

                $roles   = AAM::api()->getUser()->roles;

                if (is_null($default) && is_array($roles)) {
                    foreach($roles as $role) {
                        $default = AAM::api()->getConfig(
                            'feature.post.defaultTerm.role.' . $role, false
                        );
                        if (!empty($default)) {
                            break;
                        }
                    }
                }
            }
        }
        
        return (!empty($default) ? $default : $category);
    }
    
    /**
     * 
     * @param type $post_ID
     * @param type $post
     */
    public function savePost($post_ID, $post) {
        // first let's determine if there is any default term defined for a particular
        // post type that is not 'category'. The `category` post type is handled by
        // filterDefaultCategory method
        $taxonomies = array_filter(
            get_object_taxonomies($post), 'is_taxonomy_hierarchical'
        );
        
        foreach($taxonomies as $taxonomy) {
            // Give an opportunity to redefine default term with AAM UI
            $value = intval(AAM::api()->getUser()->getObject('taxonomy', "{$post->post_type}|{$taxonomy}")->get('default'));
            
            if (!empty($value)) {
                $terms = wp_get_object_terms($post_ID, $taxonomy);
                if (!is_array($terms) || empty($terms)) {
                    wp_set_object_terms($post_ID, $value, $taxonomy);
                }
            }
        }
    }
    
    /**
     * Undocumented function
     *
     * @param [type] $caps
     * @param [type] $cap
     * @param [type] $args
     * @return void
     */
    public function mapMetaCaps($caps, $cap, $args) {
        $objectId = (isset($args[0]) ? $args[0] : null);

        switch($cap) {
            case 'edit_term':
                $caps = $this->authorizeTermAction($caps, $objectId, 'edit');
                break;
            
            case 'delete_term':
                $caps = $this->authorizeTermAction($caps, $objectId, 'delete');
                break;
            
            case 'assign_term':
                $caps = $this->authorizeTermAction($caps, $objectId, 'assign');
                break;
            
            default:
                break;
        }
        
        return $caps;
    }
    
    /**
     * 
     * @param type $caps
     * @param type $termId
     * @param type $action
     * @return type
     */
    protected function authorizeTermAction($caps, $termId, $action) {
        $term = get_term($termId);
        
        if (is_a($term, 'WP_Term')) {
            $object = AAM::getUser()->getObject('term', "{$term->term_id}|{$term->taxonomy}");
            $area   = $this->getArea();
            
            if ($object->has('term', "{$area}.{$action}")) {
                $caps[] = 'do_not_allow';
            }
        }
        
        return $caps;
    }
    
    /**
     * 
     * @return type
     */
    protected function getArea() {
        if (defined('REST_REQUEST') && REST_REQUEST) {
            $area = 'api';
        } else {
            $area = (is_admin() ? 'backend' : 'frontend');
        }
        
        return $area;
    }
    
    /**
     * 
     * @param array $options
     * 
     * @return type
     */
    public function addPostListOptions($options) {
        return array_merge(
            $options, 
            array(
                'frontend.list_others', 'backend.list_others', 'api.list_others',
                'post|frontend.list_others', 'post|backend.list_others', 'post|api.list_others',
                'post|frontend.list', 'post|backend.list', 'post|api.list',
            )
        );
    }
    
    /**
     * 
     * @param type $result
     * @param type $property
     * @param AAM_Core_Object_Post $post
     */
    public function allowed($result, $property, AAM_Core_Object_Post $post) {
        $others = $post->has("{$property}_others");
        $user   = get_current_user_id();
        
        if ($others && (intval($post->post_author) !== $user)) {
            $option = $post->getOption();
            $result = isset($option[$property]) && empty($option[$property]);
        }
        
        return $result;
    }
    
    /**
     * 
     * @param type $args
     * @param type $taxonomy
     * @param type $objectTypes
     */
    public function registerTaxonomyArgs($args, $taxonomy, $objectTypes) {
        // first of all, lets make sure that we have capabilities defined
        $args['capabilities'] = array_merge(array(
            'manage_terms' => 'manage_categories',
            'edit_terms'   => 'manage_categories',
            'delete_terms' => 'manage_categories',
            'assign_terms' => 'edit_posts',
        ), isset($args['capabilities']) && is_array($args['capabilities']) ? $args['capabilities'] : array());
        
        // get AAM api instance
        $api = AAM::api();
        
        foreach($objectTypes as $type) {
            foreach (array_keys($args['capabilities']) as $action) {
                $capability = "{$type}_{$taxonomy}_{$action}";
                
                if ($api->capabilityExists($capability)) {
                    $args['capabilities'][$action] = $capability;
                }
            }
        }
        
        return $args;
    }
    
    /**
     * Control Post Type registration
     * 
     * @param string $post_type
     * 
     * @return void
     * 
     * @access public
     */
    public function registeredPostType($post_type, $object) {
        $type = AAM::getUser()->getObject('type', $post_type);
        $area = is_admin() ? 'backend' : 'api';

        if ($type->has('post', $area . '.add')) {
            $object->cap->create_posts = "aamp-" . uniqid();
        }
    }

    /**
     * Retrieve post terms
     * 
     * The list of terms will be also filtered if any filters are defined with 
     * ConfigPress
     *
     * @return array
     * @access private
     */
    public function retrievePostTerms($post) {
        $taxonomies = get_object_taxonomies($post);
        
        if (is_array($taxonomies) && count($taxonomies)) {
            if (!isset($this->cache[$post->ID])) {
                $this->cache[$post->ID] = $this->fetchTermsByPostIds(
                    $post->ID, $post->post_type
                );
            }

            $terms = $this->cache[$post->ID];
            
            // Critical bug fix. WordPress automatically creates a page or hierarchical
            // post type in "draft" state right after the title has been entered. This
            // creates chicken-egg problem as AAM will block any attempts to update a
            // draft post if default access to all posts is restricted to edit. That is
            // why it is important to also take in consideration all terms that have been
            // submitted on the post's edit screen
            $submitted = isset($_POST['tax_input']) ? $_POST['tax_input'] : array();
            if (!empty($submitted)) {
                foreach($this->filterTaxonomies($post->post_type, $taxonomies) as $tax) {
                    if (!empty($submitted[$tax])) {
                        $this->mergeToTerms($terms, (array)$submitted[$tax], $tax);
                    }
                }
            }
        } else {
            $terms = array();
        }
        
        return $terms;
    }

    /**
     * Undocumented function
     *
     * @param [type] $ids
     * @param [type] $postType
     * @return void
     */
    protected function fetchTermsByPostIds($ids, $postType) {
        $this->skipFilter = true;
        // Important to reset $list indexes as WP core has bug that expects
        // all taxonomies to start from index 0
        $terms = wp_get_object_terms(
            $ids, 
            $this->filterTaxonomies($postType, get_object_taxonomies($postType)), 
            array(
                'fields'                 => 'all_with_object_id',
                'update_term_meta_cache' => false
            )
        );
        $this->skipFilter = false;

        return $terms;
    }
    
    /**
     * 
     * @param type $terms
     * @param type $newTerms
     * @param type $taxonomy
     */
    protected function mergeToTerms(&$terms, $newTerms, $taxonomy) {
        foreach($newTerms as $id) {
            if ($id) {
                $has = false;
                foreach($terms as $term) {
                    if (intval($term->term_id) === intval($id)) {
                        $has = true;
                        break;
                    }
                }
                
                if (!$has) {
                    $terms[] = get_term($id, $taxonomy);
                }
            }
        }
    }
    
    /**
     * Filter list of taxonomies
     * 
     * @param string $type
     * @param array  $taxonomies
     * 
     * @return array
     * @access private
     */
    protected function filterTaxonomies($type, $taxonomies) {
        $config  = AAM::api()->getConfig("access.{$type}.taxonomy");
        $allowed = (!empty($config) ? explode(',', $config) : array());
        
        return array_values(
            count($allowed) ? array_intersect($taxonomies, $allowed) : $taxonomies
        );
    }
    
    /**
     * Filter terms
     * 
     * @param array $terms
     * 
     * @access public
     */
    public function filterTerms($terms) {
        if (is_array($terms) && !AAM::isAAM() && !$this->skipFilter) {
            if (defined('REST_REQUEST') && REST_REQUEST) {
                $area = 'api';
            } else {
                $area = (is_admin() ? 'backend' : 'frontend');
            }
            
            foreach ($terms as $i => $term) {
                if (is_a($term, 'WP_Term')) {
                    $object = AAM_PlusPackage_Base::getInstance()->initTermAccessSettings(
                            $term->term_id, $term->taxonomy
                    );
                    if ($object->has('term', $area . '.list')) {
                        unset($terms[$i]);
                    }
                }
            }
        }
        
        return $terms;
    }
    
    /**
     * Initialize term access settings
     * 
     * The Term object after initialization only fetch settings that belong to term
     * itself and any direct parent terms. However the full chain of inheritance has
     * to be triggered separately to avoid issues with cross-cut inheritance.
     * 
     * @param int    $id
     * @param string $taxonomy
     * 
     * @return AAM_Object_Term
     * 
     * @access public
     */
    public function initTermAccessSettings($id, $taxonomy) {
        $user   = AAM::getUser();
        $object = $user->getObject('term', "{$id}|{$taxonomy}");
        
        if (!$object->getOption()) { //inherit from Type default
            $tax = get_taxonomy($taxonomy);
            if (!empty($tax->object_type[0])) {
                $postType = $tax->object_type[0];
                $type     = new AAM_Object_Type($user, $postType);
                
                $object->setOption($type->getOption());
            }
        }

        if (!$object->getOption()) { //inherit from parent role terms
            $object->setOption($user->inheritFromParent('term', "{$id}|{$taxonomy}"));
        }
        
        if (!$object->getOption() && !empty($postType)) { //inherit from parent role default settings
            $object->setOption($user->inheritFromParent('type', $postType));
        }
        
        return $object;
    }
    
    /**
     * Get single instance of the class
     * 
     * This is more an alias for the bootstrap method
     * 
     * @return AAM_PlusPackage_Base
     * 
     * @access public static
     */
    public static function getInstance() {
        return self::bootstrap();
    }

    /**
     * Bootstrap the backend
     * 
     * @return AAM_PlusPackage_Base
     * 
     * @access public
     */
    public static function bootstrap() {
        if (is_null(self::$_instance)) {
            self::$_instance = new self;
        }

        return self::$_instance;
    }

}