<?php

/**
 * Copyright (C) VasylTech, LLC <vasyl@vasyltech.com>
 *
 * This program is commercial software: you are not allowed to redistribute it
 * and/or modify. Unauthorized copying of this file, via any medium is strictly
 * prohibited.
 * For any questions or concerns contact Vasyl Martyniuk <vasyl@vasyltech.com>
 */

/**
 * Term object
 * 
 * @package AAM
 * @author Vasyl Martyniuk <vasyl@vasyltech.com>
 */
class AAM_Object_Term extends AAM_Core_Object {

    /**
     *
     * @var type
     */
    private $_term = null;

    /**
     * Constructor
     *
     * @param AAM_Core_Subject $subject
     * @param Int              $termId
     * @param string           $taxonomy
     *
     * @return void
     *
     * @access public
     */
    public function __construct(AAM_Core_Subject $subject, $termId, $taxonomy = '') {
        parent::__construct($subject);

        if (is_numeric($termId)) {
            $term = get_term($termId, $taxonomy);
        } else {
            $term = get_term_by('slug', $termId, $taxonomy);
        }
        
        //make sure that we are dealing with term object
        if ($term && !is_wp_error($term)) {
            $this->setTerm($term);
            $this->initialize();
        }
    }
    
    /**
     * 
     */
    public function initialize() {
        if ($this->getTerm()) {
            $this->read();
        }
    }

    /**
     * Read the term
     * 
     * @return void
     * 
     * @access protected
     */
    protected function read() {
        $subject = $this->getSubject();
        $term    = $this->getTerm();
        $id      = $term->term_id . '|' . $term->taxonomy;
        
        $option = $subject->readOption('term', $id);
        $this->setOverwritten(!empty($option));
        
        // Check Access Policy
        if (empty($option)) {
            $stms = AAM::api()->getPolicyManager($subject)->find(
                "/^term:{$term->taxonomy}:({$term->term_id}|{$term->slug}):/",
                array('term' => $term)
            );
            $option = array();
            $compatibility = AAM::api()->getCompatibilityManager();

            foreach($stms as $key => $stm) {
                $chunks = explode(':', $key);

                if (isset($chunks[3]) && $chunks[3] === 'posts') {
                    $prefix = 'post|';
                    $action = (isset($chunks[4]) ? $chunks[4] : 'read');
                } else {
                    $prefix = 'term|';
                    $action = (isset($chunks[3]) ? $chunks[3] : 'read');
                }
                $meta = (isset($stm['Metadata']) ? $stm['Metadata'] : array());

                $option = array_merge(
                    $option,
                    $compatibility->convertPolicyAction(
                        $action,
                        $stm['Effect'] === 'deny',
                        $prefix,
                        ($action === 'read' ? $meta : array())
                    )
                );
            }
        }
        
        if (empty($option) && !empty($term->parent)) { //try to get any parent settings
            $option = $this->inheritFromParent($term->parent, $term->taxonomy);
        }
        
        $this->setOption($option);
    }
    
    /**
     * Inherit options from parent
     * 
     * Drill up to by term's hierarchy and find the first term that has access
     * options defined.
     * 
     * @param int    $term_id
     * @param string $taxonomy
     * 
     * @return array
     * 
     * @access protected
     */
    protected function inheritFromParent($term_id, $taxonomy) {
        $object = new AAM_Object_Term($this->getSubject(), $term_id, $taxonomy);
        if ($object->getTerm()) {
            $access = $object->getOption();
            if (empty($access) && $object->getTerm()->parent) {
                $this->inheritFromParent(
                        $object->getTerm()->parent,
                        $object->getTerm()->taxonomy
                );
            }
        } else {
            $access = array();
        }

        return $access;
    }
    
    /**
     * Save options
     * 
     * @param string  $property
     * @param boolean $checked
     * 
     * @return boolean
     * 
     * @access public
     */
    public function save($property, $checked) {
        $termId  = $this->getTerm()->term_id . '|' . $this->getTerm()->taxonomy;
        
        $option            = $this->getOption();
        $option[$property] = $checked;
        
        $result = $this->getSubject()->updateOption($option, 'term', $termId);
        
        if ($result) {
            $this->setOption($option);
        }
        
        return $result;
    }
    
    /**
     * 
     * @return type
     */
    public function reset() {
        $id  = $this->getTerm()->term_id . '|' . $this->getTerm()->taxonomy;
        
        return $this->getSubject()->deleteOption('term', $id);
    }

    /**
     * Check if property is selected
     * 
     * @param string $group
     * @param string $action
     * 
     * @return boolean
     * 
     * @access public
     */
    public function has($group, $action) {
        $option = $this->getGroup($group);
        
        return !empty($option[$action]);
    }
    
    /**
     * 
     * @param type $group
     * @param type $action
     * @return type
     */
    public function get($group, $action) {
        $option = $this->getGroup($group);
        
        return isset($option[$action]) ? $option[$action] : null;
    }
    
    /**
     * Get group of options
     * 
     * Term object stores property in a bit different way just because it has
     * to hold both child posts and term default settings. That is why the property
     * has a different format "[post|term]|property". It consists of two parts 
     * separated by the pipe "|" where first part is either "post" or "term" and 
     * another part is an actual access property (e.g. frontend.list)
     * 
     * @param string $name
     * 
     * @return array
     * 
     * @access public
     */
    public function getGroup($name) {
        $split = array();
        
        foreach($this->getOption() as $key => $value) {
            list($group, $property) = explode('|', $key);
            $split[$group][$property] = $value;
        }
        
        //Very important. If category group is set but not post group than inheritance
        //for post will not work, so make sure that this case is covered
        if ($this->isOverwritten()) {
            $default = array('default' => false);
        } else {
            $default = array();
        }
        
        return (!empty($split[$name]) ? $split[$name] : $default);
    }
    
    /**
     * Set term
     * 
     * @param stdClass $term
     * 
     * @return void
     * 
     * @access public
     */
    public function setTerm($term) {
        $this->_term = $term;
    }

    /**
     * Get term
     * 
     * @return stdClass
     * 
     * @access public
     */
    public function getTerm() {
        return $this->_term;
    }
    
    /**
     * 
     * @param type $external
     * @return type
     */
    public function mergeOption($external) {
        return AAM::api()->mergeSettings($external, $this->getOption(), 'term');
    }
    
}