<?php

/**
 * Copyright (C) VasylTech, LLC <vasyl@vasyltech.com>
 *
 * This program is commercial software: you are not allowed to redistribute it
 * and/or modify. Unauthorized copying of this file, via any medium is strictly
 * prohibited.
 * For any questions or concerns contact Vasyl Martyniuk <vasyl@vasyltech.com>
 */

/**
 * Post type object
 * 
 * @package AAM
 * @author Vasyl Martyniuk <vasyl@vasyltech.com>
 */
class AAM_Object_Type extends AAM_Core_Object {

    /**
     * Post type
     * 
     * @var string 
     * 
     * @access private
     */
    private $_postType;

    /**
     * Constructor
     *
     * @param AAM_Core_Subject $subject
     * @param string           $postType
     *
     * @return void
     *
     * @access public
     */
    public function __construct(AAM_Core_Subject $subject, $postType) {
        parent::__construct($subject);

        $option = $subject->readOption('type', $postType);
        $this->setOverwritten(!empty($option)); 
        
        // Check Access Policy
        if (empty($option)) {
            $stms = AAM::api()->getPolicyManager($subject)->find(
                "/^posttype:{$postType}:/",
                array('postType' => $postType)
            );
            $option = array();
            $compatibility = AAM::api()->getCompatibilityManager();

            foreach($stms as $key => $stm) {
                $chunks = explode(':', $key);
                $action = (isset($chunks[3]) ? $chunks[3] : 'read');
                $meta   = (isset($stm['Metadata']) ? $stm['Metadata'] : array());

                $option = array_merge(
                    $option,
                    $compatibility->convertPolicyAction(
                        $action,
                        $stm['Effect'] === 'deny',
                        ($chunks[2] === 'posts' ? 'post|' : 'term|'),
                        ($action === 'read' ? $meta : array())
                    )
                );
            }
        }

        $this->setOption($option);
        $this->setPostType($postType);
    }

    /**
     * Save settings
     * 
     * @param string  $property
     * @param boolean $checked
     * 
     * @return boolean
     * 
     * @access public
     */
    public function save($property, $checked) {
        $subject = $this->getSubject();
        
        $option            = $this->getOption();
        $option[$property] = $checked;
        
        $result = $subject->updateOption($option, 'type', $this->getPostType());
        
        if ($result) {
            $this->setOption($option);
        }
        
        return $result;
    }
    
    /**
     * 
     * @return type
     */
    public function reset() {
        return $this->getSubject()->deleteOption('type', $this->getPostType());
    }

    /**
     * Check if property is selected
     * 
     * @param string $group
     * @param string $action
     * 
     * @return boolean
     * 
     * @access public
     */
    public function has($group, $action) {
        $option = $this->getGroup($group);
        
        return !empty($option[$action]);
    }
    
    /**
     * Get group of options
     * 
     * Post type object stores property in a bit different way just because it has
     * to hold both posts and terms default settings. That is why the property has a
     * different format "[post|term]|property". It consists of two parts separated by
     * the pipe "|" where first part is either "post" or "term" and another part is
     * an actual access property (e.g. frontend.list, backend.edit etc.)
     * 
     * @param string  $name
     * @param boolean $prefix Keep the full property name or not
     * 
     * @return array
     * 
     * @access public
     */
    public function getGroup($name, $prefix = false) {
        $split = array();
        
        foreach($this->getOption() as $key => $value) {
            list($group, $property) = explode('|', $key);
            $split[$group][($prefix ? $key : $property)] = $value;
        }
        
        //Very important. If category group is set but not post group than inheritance
        //for post will not work, so make sure that this case is covered
        if ($this->isOverwritten()) {
            $default = array('default' => false);
        } else {
            $default = array();
        }
        
        return (!empty($split[$name]) ? $split[$name] : $default);
    }

    /**
     * Set current post type
     * 
     * @param string $postType
     * 
     * @return void
     * 
     * @access public
     */
    public function setPostType($postType) {
        $this->_postType = $postType;
    }

    /**
     * Get current post type
     * 
     * @return string
     * 
     * @access public
     */
    public function getPostType() {
        return $this->_postType;
    }
    
    /**
     * 
     * @param type $external
     * @return type
     */
    public function mergeOption($external) {
        return AAM::api()->mergeSettings($external, $this->getOption(), 'type');
    }

}