<?php

/**
 * Copyright (C) VasylTech, LLC <vasyl@vasyltech.com>
 *
 * This program is commercial software: you are not allowed to redistribute it
 * and/or modify. Unauthorized copying of this file, via any medium is strictly
 * prohibited.
 * For any questions or concerns contact Vasyl Martyniuk <vasyl@vasyltech.com>
 */

/**
 * AAM Plus Package Visibility Initializer
 *
 * @package AAM
 * @author Vasyl Martyniuk <vasyl@vasyltech.com>
 * @since v3.9
 */
class AAM_PlusPackage_Visibility_Initializer {

    /**
     * Single instance of itself
     * 
     * @var AAM_PlusPackage_Visibility_Initializer 
     */
    protected static $instance = null;

    /**
     * 
     */
    protected function __construct() {}

    /**
     * 
     * @param AAM_Core_Object $object
     */
    public function execute(AAM_Core_Object $object) {
        $subject = $object->getSubject();

        if ($subject::UID == 'user') {
            $this->initializeUserVisibility($object);
        } elseif ($subject::UID == 'role') {
            $this->initializeRoleVisibility($object);
        } elseif ($subject::UID == 'default') {
            $this->initializeDefaultVisibility($object);
        } elseif ($subject::UID == 'visitor') {
            $this->initializeVisitorVisibility($object);
        }
        
        if (AAM::api()->getConfig('core.settings.inheritParentPost', false)) {
            if ($object->has('post')) {
                $option = $object->getOption();
                $this->initializePostHierarchy($object, $option['post']);
            }
        }
        
        // Read all the settings from the Access & Security Policies
        $area = $this->getArea();
        $stms = AAM::api()->getPolicyManager($subject)->find(
            '/:(list|listtoothers)$/'
        );
        
        foreach($stms as $key => $stm) {
            $chunks = explode(':', $key);

            switch($chunks[0]) {
                case 'posttype':
                    $action = ($chunks[3] === 'listtoothers' ? 'list_others' : 'list');
                    $prefix = ($chunks[2] === 'posts' ? 'post' : 'term');

                    $object->pushOptions(
                        'type', 
                        $chunks[1], 
                        array(
                            "{$prefix}|{$area}.{$action}" => ($stm['Effect'] === "deny" ? 1 : 0)
                        )
                    );
                    break;

                case 'term':
                    if (is_numeric($chunks[2])) {
                        $termId = $chunks[2];
                    } else {
                        $term   = get_term_by('slug', $chunks[2], $chunks[1]);
                        $termId = is_a($term, 'WP_Term') ? $term->term_id : 0;
                    }

                    if (!empty($termId)) {
                        // Term:category:12:posts (all child posts)
                        if (isset($chunks[3]) && $chunks[3] === 'posts') {
                            $action = ($chunks[4] === 'listtoothers' ? 'list_others' : 'list');

                            $object->pushOptions(
                                'term', 
                                "{$termId}|{$chunks[1]}", 
                                array(
                                    "post|{$area}.{$action}" => ($stm['Effect'] === "deny" ? 1 : 0)
                                )
                            );
                        } else {
                            // TODO: Not too obvious
                            $action = ($chunks[3] === 'listtoothers' ? 'list_others' : 'list');
                            $object->pushOptions(
                                'term', 
                                "{$termId}|{$chunks[1]}", 
                                array(
                                    "term|{$area}.{$action}" => ($stm['Effect'] === "deny" ? 1 : 0)
                                )
                            );
                        }
                    }
                    break;

                default:
                    break;
            }
        }
    }
    
     /**
     * 
     * @return type
     */
    protected function getArea() {
        if (defined('REST_REQUEST') && REST_REQUEST) {
            $area = 'api';
        } else {
            $area = (is_admin() ? 'backend' : 'frontend');
        }
        
        return $area;
    }

    /**
     * 
     * @global type $wpdb
     * @param type $object
     */
    protected function initializeUserVisibility($object) {
        global $wpdb;

        $uid = $object->getSubject()->ID;

        $query  = "SELECT meta_key, meta_value FROM {$wpdb->usermeta} WHERE ";
        $query .= "(user_id = %d) AND ((meta_key LIKE %s) OR (meta_key LIKE %s))";

        if ($wpdb->query($wpdb->prepare($query, $uid, '%aam_term%', '%aam_type%'))) {
            $terms = array();
            
            foreach ($wpdb->last_result as $row) {
                $chunks = $this->parseUserOptionName($row->meta_key);
                if ($chunks) {
                    $id = explode('|', $chunks['id']);

                    $settings = $object->pushOptions(
                        $chunks['type'], $chunks['id'], maybe_unserialize($row->meta_value)
                    );

                    if ($chunks['type'] == 'term') {
                        $terms[$chunks['id']] = array_merge($id, array($settings));
                    }
                }
            }
            
            if (!empty($terms)) {
                $this->initializeTermHierarchy($object, $terms);
            }
        }
        
        
    }

    /**
     * 
     * @global type $wpdb
     * @param type $object
     */
    protected function initializeRoleVisibility($object) {
        global $wpdb;

        $query = "SELECT option_name, option_value FROM {$wpdb->options} WHERE ";
        $query .= "(option_name LIKE %s)";

        $role = $object->getSubject()->getId();

        if ($wpdb->query($wpdb->prepare($query, 'aam_%_role_' . $role))) {
            $terms = array();
            
            foreach ($wpdb->last_result as $row) {
                $chunks = $this->parseRoleOptionName($row->option_name, $role);
                if ($chunks) {
                    $id = explode('|', $chunks['id']);

                    $settings = $object->pushOptions(
                        $chunks['type'], $chunks['id'], maybe_unserialize($row->option_value)
                    );

                    if ($chunks['type'] == 'term') {
                        $terms[$chunks['id']] = array_merge($id, array($settings));
                    }
                }
            }
            
            if (!empty($terms)) {
                $this->initializeTermHierarchy($object, $terms);
            }
        }
    }

    /**
     * 
     * @global type $wpdb
     * @param type $object
     */
    protected function initializeDefaultVisibility($object) {
        global $wpdb;

        $query = "SELECT option_name, option_value FROM {$wpdb->options} WHERE ";
        $query .= "(option_name LIKE %s)";

        if ($wpdb->query($wpdb->prepare($query, 'aam_%_default'))) {
            $terms = array();
            
            foreach ($wpdb->last_result as $row) {
                $chunks = $this->parseDefaultOptionName($row->option_name);
                if ($chunks) {
                    $id = explode('|', $chunks['id']);

                    $settings = $object->pushOptions(
                        $chunks['type'], $chunks['id'], maybe_unserialize($row->option_value)
                    );

                    if ($chunks['type'] == 'term') {
                        $terms[$chunks['id']] = array_merge($id, array($settings));
                    }
                }
            }
            
            if (!empty($terms)) {
                $this->initializeTermHierarchy($object, $terms);
            }
        }
        
        $query  = "SELECT pm.`post_id`, pm.`meta_value`, p.`post_type` FROM {$wpdb->postmeta} AS pm ";
        $query .= "LEFT JOIN {$wpdb->posts} AS p ON (pm.`post_id` = p.ID) ";
        $query .= "WHERE pm.`meta_key` = %s";

        if ($wpdb->query($wpdb->prepare($query, 'aam-post-access-default'))) {
            $terms = array();
            
            foreach ($wpdb->last_result as $row) {
                
                if (!$object->has('post', $row->post_id)) {
                    $object->pushOptions(
                            'post', $row->post_id . '|' . $row->post_type, maybe_unserialize($row->meta_value)
                    );
                }
            }
        }
    }
    
    /**
     * 
     * @global type $wpdb
     * @param type $object
     */
    protected function initializeVisitorVisibility($object) {
        global $wpdb;

        $query = "SELECT option_name, option_value FROM {$wpdb->options} WHERE ";
        $query .= "(option_name LIKE %s) OR (option_name LIKE %s)";

        if ($wpdb->query($wpdb->prepare($query, 'aam_visitor_term%', 'aam_visitor_type%'))) {
            $terms = array();
            
            foreach ($wpdb->last_result as $row) {
                $chunks = $this->parseVisitorOptionName($row->option_name);
                if ($chunks) {
                    $id = explode('|', $chunks['id']);

                    $settings = $object->pushOptions(
                        $chunks['type'], $chunks['id'], maybe_unserialize($row->option_value)
                    );

                    if ($chunks['type'] === 'term') {
                        $terms[$chunks['id']] = array_merge($id, array($settings));
                    }
                }
            }
            
            if (!empty($terms)) {
                $this->initializeTermHierarchy($object, $terms);
            }
        }
    }
    
    /**
     * 
     * @param type $name
     * @return type
     */
    protected function parseUserOptionName($name) {
        $response = false;
        
        if (preg_match('/aam_(term|type)_([\w_\-\|]+)$/', $name, $match)) {
            $response = array(
                'type' => $match[1],
                'id'   => $match[2]
            );
        }
        
        return $response;
    }
    
    /**
     * 
     * @param type $name
     * @return type
     */
    protected function parseRoleOptionName($name, $id) {
        $response = false;
        
        if (preg_match('/^aam_(term|type)_([\w_\-\|]+)_role_(' . $id . ')$/', $name, $match)) {
            $response = array(
                'type' => $match[1],
                'id'   => $match[2]
            );
        }
        
        return $response;
    }
    
    /**
     * 
     * @param type $name
     * @return type
     */
    protected function parseDefaultOptionName($name) {
        $response = false;
        
        if (preg_match('/^aam_(term|type)_([\w\-\|]+)_default$/', $name, $match)) {
            // Cover the scenario when there is a custom role "default"
            if (!preg_match('/_role_default$/', $name) 
                    || !$this->getRoles()->is_role('default')) {
                $response = array(
                    'type' => $match[1],
                    'id'   => $match[2]
                );
            }
        }
        
        return $response;
    }
    
    /**
     * 
     * @param type $name
     * @return type
     */
    protected function parseVisitorOptionName($name) {
        $response = false;
        
        if (preg_match('/^aam_visitor_(term|type)_([\w_\-\|]+)$/', $name, $match)) {
            $response = array(
                'type' => $match[1],
                'id'   => $match[2]
            );
        }
        
        return $response;
    }
    
    /**
     * Get role list
     * 
     * @global WP_Roles $wp_roles
     * 
     * @return WP_Roles
     */
    protected static function getRoles() {
        global $wp_roles;
        
        if (function_exists('wp_roles')) {
            $roles = wp_roles();
        } elseif(isset($wp_roles)) {
            $roles = $wp_roles;
        } else {
            $roles = new WP_Roles();
        }
        
        return $roles;
    }
    

    /**
     * 
     * @param type $object
     * @param type $terms
     */
    protected function initializeTermHierarchy($object, $terms) {
        foreach($terms as $term) {
            $children = get_term_children($term[0], $term[1]);
            $settings = $term[2];
            if (is_array($children)) {
                foreach($children as $child) {
                    $id = "{$child}|{$term[1]}";
                    if (is_numeric($child)) {
                        $settings = (isset($terms[$id]) ? $terms[$id][2] : $settings);
                        $object->pushOptions('term', $id, $settings);
                    }
                }
            }
        }
    }
    
    /**
     * 
     * @param type $object
     * @param type $posts
     */
    protected function initializePostHierarchy($object, $posts) {
        foreach($posts as $id => $settings) {
            $this->initializePostChildren($id, $settings, $object, $posts);
        }
    }
    
    /**
     * 
     * @global type $wpdb
     * @param type $id
     * @param type $settings
     * @param type $object
     * @param type $posts
     */
    protected function initializePostChildren($id, $settings, $object, $posts) {
        global $wpdb;
        
        $chunks = explode('|', $id);
        
        $query  = "SELECT `ID`, `post_type` FROM {$wpdb->posts} ";
        $query .= "WHERE post_parent = %d AND post_status NOT IN ('" . implode("','", get_post_stati(array('exclude_from_search' => true))) . "')";

        if ($wpdb->query($wpdb->prepare($query, $chunks[0]))) {
            foreach($wpdb->last_result as $child) {
                $childId = $child->ID . '|' . $child->post_type;
                
                $object->pushOptions(
                    'post', 
                    $childId, 
                    (isset($posts[$childId]) ? $posts[$childId] : $settings)
                );
                
                $this->initializePostChildren(
                    $childId, 
                    (isset($posts[$childId]) ? $posts[$childId] : $settings), 
                    $object, 
                    $posts
                );
            }
        }
    }

    /**
     * Bootstrap the initializer
     * 
     * @return AAM_PlusPackage_Visibility_Initializer
     */
    public static function bootstrap() {
        if (is_null(self::$instance)) {
            self::$instance = new self;
        }
        
        return self::$instance;
    }

}